# CloudFormation does not currently support and probably will never
# 
# June 2022
# 
# https://github.com/bridgecrewio/checkov/blob/master/checkov/terraform/checks/resource/aws/NeptuneClusterInstancePublic.py
#
# check_id: CKV_AWS_102
